#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai


__license__   = 'GPL v3'
__copyright__ = '2021, Bradley Lang'
__docformat__ = 'restructuredtext en'

try:
    load_translations()
except NameError:
    pass # load_translations() added in calibre 1.9

# these are required for the test connection button
import json
import urllib

import logging
logger = logging.getLogger(__name__)

try:
    from qt.core import QWidget, QVBoxLayout, QHBoxLayout, QDialogButtonBox, QLabel, QLineEdit, QIntValidator, QPushButton, QMessageBox, QGroupBox, QComboBox, QCheckBox
except ImportError:
    from PyQt5.Qt import QWidget, QVBoxLayout, QHBoxLayout, QDialogButtonBox, QLabel, QLineEdit, QIntValidator, QPushButton, QMessageBox, QGroupBox, QComboBox, QCheckBox

from calibre.utils.config import JSONConfig


# This is where all preferences for this plugin will be stored
# Remember that this name (i.e. plugins/librarything_match) is also
# in a global namespace, so make it as unique as possible.
# You should always prefix your config file name with plugins/,
# so as to ensure you dont accidentally clobber a calibre config file
prefs = JSONConfig('plugins/librarything_match')

# Set defaults
prefs.defaults['librarything_user'] = 'sample'                          # Exists in LibraryThing.
prefs.defaults['librarything_key'] = 'not required'                     # Will disable this field in the UI.
prefs.defaults['librarything_coll'] = _('All collections')              # This is a "pseudo-collection" for all books.
prefs.defaults['librarything_taglist'] = ''                             # Comma-sep list of LT tags for which to search.
prefs.defaults['librarything_sort'] = 'author,title'                    # Author then title.
prefs.defaults['max_results'] = '100'                                   # This needs to be big enough for all LT items, i.e. "All collections".
prefs.defaults['librarything_id_column'] = '#librarything_id'           # Custom column to store ltid
prefs.defaults['librarything_rating_column'] = '#librarything_rating'   # Custom column to store lt rating
prefs.defaults['probable_match_threshold'] = '90'                       # What confidence level do we consider a good match?
prefs.defaults['use_calibre_identifier'] = False                        # If True will copy the librarything_id into a calibre identifier ltid.
prefs.defaults['logging_level'] = 'WARNING'                             # CRITICAL, ERROR, WARNING, INFO or DEBUG (or NOTSET)
prefs.defaults['logging_logfile'] = ''                                  # Leave blank to log to console.
prefs.defaults['ignore_list'] = ''                                      # LT books to permanently ignore.


# LibraryThing sorting options. (key used in URL, value displayed in QComboBox.)
sort_options = {
    "random"        : _("Random"),
    "entry"         : _("Entry date order"),
    "entry_REV"     : _("Entry date descending"),
    "title"         : _("Title (A-Z)"),
    "title_REV"     : _("Title descending (Z-A)"),
    "author"        : _("Author (Last, First)"),
    "author_REV"    : _("Author (Last, First) desc"),
    "author,title"  : _("Author then Title")
}

class ConfigWidget(QWidget):

    def __init__(self):
        QWidget.__init__(self)

        self.l = QVBoxLayout()
        self.setLayout(self.l)

        groupBox = QGroupBox(_("LibraryThing connection"))
        self.l.addWidget(groupBox)

        groupBox_hlayout = QHBoxLayout()
        groupBox.setLayout (groupBox_hlayout)


        #accelerators: u k c o m g t i r p a d

        vboxlabels = QVBoxLayout()
        groupBox_hlayout.addLayout(vboxlabels)

        # LT User label
        usrlabel = QLabel(_('LibraryThing &user:'))
        vboxlabels.addWidget(usrlabel)

        # LT User Key label
        keylabel = QLabel(_('LibraryThing &key:'))
        vboxlabels.addWidget(keylabel)

        # LT Collection label
        colllabel = QLabel(_('LibraryThing &collection:'))
        vboxlabels.addWidget(colllabel)

        # LT taglist
        taglabel = QLabel(_('LibraryThing t&aglist:'))
        vboxlabels.addWidget(taglabel)

        # LT Sort label
        sortlabel = QLabel(_('LT sort &order:'))
        vboxlabels.addWidget(sortlabel)

        # Max Results label
        resultslabel = QLabel(_('&Max results:'))
        vboxlabels.addWidget(resultslabel)
        


        vboxeditors = QVBoxLayout()
        groupBox_hlayout.addLayout(vboxeditors)

        # LT User edit
        self.ltusr = QLineEdit(self)
        self.ltusr.setText(prefs['librarything_user'])
        vboxeditors.addWidget(self.ltusr)
        usrlabel.setBuddy(self.ltusr)

        # LT User Key edit
        self.ltkey = QLineEdit(self)
        self.ltkey.setText(prefs['librarything_key'])
        vboxeditors.addWidget(self.ltkey)
        keylabel.setBuddy(self.ltkey)
        self.ltkey.setEnabled(False)

        # LT Collection edit
        self.ltcoll = QLineEdit(self)
        self.ltcoll.setText(prefs['librarything_coll'])
        vboxeditors.addWidget(self.ltcoll)
        colllabel.setBuddy(self.ltcoll)
        
        # LT taglist edit
        self.lttaglist = QLineEdit(self)
        self.lttaglist.setText(prefs['librarything_taglist'])
        vboxeditors.addWidget(self.lttaglist)
        taglabel.setBuddy(self.lttaglist)

        # LT sort combo
        self.ltsort = QComboBox(self)
        vboxeditors.addWidget(self.ltsort)
        sortlabel.setBuddy(self.ltsort)
        for k, v in sort_options.items():
            self.ltsort.addItem(v)
        default_value = sort_options[prefs['librarything_sort']]
        self.ltsort.setCurrentIndex(self.ltsort.findText(default_value))

        # Max Results edit
        self.ltmaxresults = QLineEdit(self)
        validator = QIntValidator(1, 10000, self)
        # the  lineedit SHOULD only accept integers between 1 and 10,000, but actually accepts 99,999!
        self.ltmaxresults.setValidator(validator)
        self.ltmaxresults.setText(prefs['max_results'])
        vboxeditors.addWidget(self.ltmaxresults)
        resultslabel.setBuddy(self.ltmaxresults)        
        


        button_layout = QHBoxLayout()
        self.l.addLayout(button_layout)

        # put these outside the groupbox, coz it's too f'n hard to layout otherwise.
        self.clear_ignore_button = QPushButton(_('Clear i&gnore list...'), self)
        self.clear_ignore_button.clicked.connect(self.clear_ignore_list)
        self.clear_ignore_button.setToolTip(_("Clear the list of LT books permanently ignored by LT Match"))
        button_layout.addWidget(self.clear_ignore_button)

        button_layout.addStretch(1)

        self.test_button = QPushButton(_('&Test connection'), self)
        self.test_button.clicked.connect(self.test)
        url = "https://www.librarything.com/api_getdata.php?userid={user}&max={max}&tagList={taglist}&showCollections=1&showRatings=1&showTags=0&booksort={order}&responseType=json"
        self.test_button.setToolTip(url)
        button_layout.addWidget(self.test_button)


        self.l.addStretch(1)
        options_layout = QHBoxLayout()        
        self.l.addLayout(options_layout)

        options_labels = QVBoxLayout()
        options_layout.addLayout(options_labels)
        
        # LTID label
        lt_id_label = QLabel(_('LibraryThing &ID column:'))
        options_labels.addWidget(lt_id_label)
        
        # LT rating label
        lt_rating_label = QLabel(_('LibraryThing &rating column:'))
        options_labels.addWidget(lt_rating_label)
        
        # Threshold label
        threshold_label = QLabel(_('&Probable match threshold:'))
        options_labels.addWidget(threshold_label)
        
        # Use Calibre Identifier label
        calid_label = QLabel(_('Calibre I&dentifier:'))
        options_labels.addWidget(calid_label)
        
        options_objects = QVBoxLayout()
        options_layout.addLayout(options_objects)
        
        # LTID edit
        self.ltid_edit = QLineEdit(self)
        self.ltid_edit.setText(prefs['librarything_id_column'])
        options_objects.addWidget(self.ltid_edit)
        lt_id_label.setBuddy(self.ltid_edit)


        # LT rating edit
        self.ltrating_edit = QLineEdit(self)
        self.ltrating_edit.setText(prefs['librarything_rating_column'])
        options_objects.addWidget(self.ltrating_edit)
        lt_rating_label.setBuddy(self.ltrating_edit)


        # Threshold edit
        self.threshold_edit = QLineEdit(self)
        validator = QIntValidator(1, 100, self)
        # the  lineedit should only accept integers between 1 and 100!
        self.threshold_edit.setValidator(validator)
        self.threshold_edit.setText(prefs['probable_match_threshold'])
        options_objects.addWidget(self.threshold_edit)
        threshold_label.setBuddy(self.threshold_edit)

        # Use Calibre Identifier check box
        self.calid_checkbox = QCheckBox(_("Use calibre identifier"), self)
        self.calid_checkbox.setToolTip(_("When saving matches, copy the librarything_id to calibre identifiers (ltid:) to be used when downloading metadata"))
        self.calid_checkbox.setChecked(prefs['use_calibre_identifier'])
        options_objects.addWidget(self.calid_checkbox)
        calid_label.setBuddy(self.calid_checkbox)

        
    def save_settings(self):
    
        # ignore_list is not a user-editable setting.
        prefs['librarything_user'] = self.ltusr.text()
        prefs['librarything_key'] = self.ltkey.text()
        if self.ltcoll.text() == "":
            prefs['librarything_coll'] = _("All collections")
        else:
            prefs['librarything_coll'] = self.ltcoll.text()
        prefs['librarything_taglist'] = self.lttaglist.text()
        idx = self.ltsort.currentIndex()
        sort_key = list(sort_options)[idx]
        prefs['librarything_sort'] = sort_key
        prefs['max_results'] = self.ltmaxresults.text()
        prefs['librarything_id_column'] = self.ltid_edit.text()
        prefs['librarything_rating_column'] = self.ltrating_edit.text()
        prefs['probable_match_threshold'] = self.threshold_edit.text()
        prefs['use_calibre_identifier'] = self.calid_checkbox.isChecked()


    def test(self):

        url = "https://www.librarything.com/api_getdata.php?userid={user}&max={max}&tagList={taglist}&showCollections=1&showRatings=1&showTags=0&booksort={order}&responseType=json"
        idx = self.ltsort.currentIndex()
        sort_key = list(sort_options)[idx]
        url = url.format(
                            user = self.ltusr.text(),
                            max = self.ltmaxresults.text(),
                            taglist = urllib.parse.quote_plus(self.lttaglist.text()),
                            order = sort_key
                        )

        text = ""
        try:
            response = urllib.request.urlopen(url)
            logger.info ("response.msg %s", response.msg)
            logger.info ("response.status %s", response.status)

            data = json.loads(response.read())
            if data["books"]:
                text = _("Connection successful. Retrieved {num} books.")
                text = text.format(num = len(data["books"]))
                logger.info(text)
            else:
                text = _("Connection successful, but no books returned. Please double-check collection and taglist for the coRRect cASe.")
                text = text.format(num = len(data["books"]))
                logger.info(text)            
            QMessageBox.about(self, _('Test connection...'), text)

        except Exception as e:
            text = _("Connection failed.")
            logger.error(_("Connection failed."))
            msgBox = QMessageBox(self)
            msgBox .setWindowTitle("LibraryThing Match")
            msgBox.setText(_("Can't connect to LibraryThing due to an error."))
            msgBox.setInformativeText(_("This could be due to an incorrect username. See error details below."))
            msgBox.setDetailedText(repr(e))
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.exec_()


    def clear_ignore_list(self):

        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("LibraryThing Match")
        msgBox.setText(_("This will clear the list of LibraryThing books that LT Match permanently ignores."))
        msgBox.setInformativeText(_("Individual items can be removed by editing the JSON config file.\n\n"
                                  "Do you wish to continue?"))
        msgBox.setIcon(QMessageBox.Question)
        msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        msgBox.setDefaultButton(QMessageBox.Ok)
        ret = msgBox.exec_()
        if ret == QMessageBox.Ok:
            logger.info (_("clear_ignore_list: Clearing ignore list"))
            prefs['ignore_list'] = ''
