#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai


__license__   = 'GPL v3'
__copyright__ = '2021, Bradley Lang'
__docformat__ = 'restructuredtext en'

try:
    load_translations()
except NameError:
    pass # load_translations() added in calibre 1.9

import json
import urllib
import urllib.parse
import html
from calibre_plugins.librarything_match.config import prefs
import logging

from calibre.utils.localization import get_lang

LTID_NAME = 'ltid'

if prefs['logging_logfile'] != '':
    logging.basicConfig(
                filename=prefs['logging_logfile'],
                filemode='w',
                level=prefs['logging_level'],
                format="%(levelname)s: %(name)s - %(message)s"
                )
else:
    logging.basicConfig(
                level=prefs['logging_level'],
                format="%(levelname)s: %(name)s - %(message)s"
                )
logger = logging.getLogger(__name__)


try:
    from qt.core import QApplication, QCursor, Qt, QVBoxLayout, QHBoxLayout, QPushButton, QMessageBox, QLabel, QSizePolicy, QCheckBox, QProgressDialog, QTimer, QTableWidget, QFont
except ImportError:
    from PyQt5.Qt import QApplication, QCursor, Qt, QVBoxLayout, QHBoxLayout, QPushButton, QMessageBox, QLabel, QSizePolicy, QCheckBox, QProgressDialog, QTimer, QTableWidget, QFont

from calibre_plugins.librarything_match.widgets import LTMatchTableWidget, LTMatchBookWidget, LTMatchStatusWidget
from calibre_plugins.librarything_match.windows import SizePersistedDialog, ViewDetails

from calibre_plugins.librarything_match.fuzzywuzzy import fuzz


class MatchDialog(SizePersistedDialog):

    def __init__(self, gui, icon, do_user_config):

        SizePersistedDialog.__init__(self, gui, 'LTMatchDialog')
        self.gui = gui
        self.do_user_config = do_user_config
        self.m = self.gui.library_view.model()

        self.setWindowTitle('LibraryThing Match')
        self.setWindowIcon(icon)

        # LibraryThing custom columns. (Check existence in singleshot timer after showEvent.)
        self.lt_id_col = prefs['librarything_id_column']
        self.lt_rating_col = prefs['librarything_rating_column']

        # keep track of book_ids list in an instance variable.
        self.book_ids = []

        # hold on to the LT books as well (this is a dict built from the JSON data, keyed on librarything_id).
        self.lt_collection = {}

        # keep a dictionary of matched books, which will be key:librarything_id, value: tuple(calibre_id, row)
        self.matched = {}
        
        # keep another dict of possible matches for when we auto-match, key:lt_id, value: tuple(calibre_id, row, percent).
        self.possible_match = {}

        # a list of LT IDs to ignore.
        self.ignore_list = []

        # The current database shown in the GUI
        # db is an instance of the class LibraryDatabase from db/legacy.py
        # This class has many, many methods that allow you to do a lot of
        # things. For most purposes you should use db.new_api, which has
        # a much nicer interface from db/cache.py
        self.db = gui.current_db

        self.l = QVBoxLayout()
        self.setLayout(self.l)

        caption_layout = QHBoxLayout()
        self.l.addLayout(caption_layout)

        # User should be aware if all books are displayed, or a subset.
        cal_label = _('<b>calibre</b> // e-book list')
        caption_layout.addWidget(QLabel(cal_label))

        caption_layout.addStretch()

        self.lt_label = QLabel()
        self.update_lt_caption()
        caption_layout.addWidget(self.lt_label)

        lists_layout = QHBoxLayout()
        self.l.addLayout(lists_layout)

        self.calibre_table = LTMatchTableWidget(self)
        self.calibre_table.verticalScrollBar().valueChanged.connect(self.scrolled_cal_list)
        self.calibre_table.doubleClicked.connect(self.show_calibre_details_popup)
        lists_layout.addWidget(self.calibre_table, 10)

        # We're hiding the header of the "status" list, so need to put it in its own layout.
        divider_layout = QVBoxLayout()
        lists_layout.addLayout(divider_layout)

        # use a regular QTableWidget for the status table, as we don't need the overhead of a LTMatchTableWidget
        self.status_table = QTableWidget(self)
        self.status_table.setMaximumWidth(100)
        self.status_table.setMinimumWidth(100)
        self.status_table.verticalHeader().setVisible(False)        
        self.status_table.setAlternatingRowColors(False)
        self.status_table.setEditTriggers(QTableWidget.NoEditTriggers)
        self.status_table.setSortingEnabled(False)
        self.status_table.setColumnCount(1)
        self.status_table.horizontalHeader().hide()
        self.status_table.verticalScrollBar().valueChanged.connect(self.scrolled_status_list)
        self.status_table.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff) # still responds to scroll events
        # Transparent background.
        self.status_table.setStyleSheet("background-color: transparent;")

        # since we're hiding the header, move it down by the height of the other headers,
        # less 9 for the default padding, plus one for something.
        # divider_layout.addSpacing(self.lt_table.verticalHeader().height()) - 8) # NOT WORKING
        divider_layout.addSpacing(20)
        divider_layout.addWidget(self.status_table)

        self.lt_table = LTMatchTableWidget(self)
        self.lt_table.verticalScrollBar().valueChanged.connect(self.scrolled_lt_list)
        self.lt_table.doubleClicked.connect(self.show_lt_details_popup)
        lists_layout.addWidget(self.lt_table, 10)

        # buttons
        button_layout = QHBoxLayout()
        self.l.addLayout(button_layout)

        # about
        self.about_button = QPushButton('?', self)
        self.about_button.setFont(QFont('Arial', 14, QFont.Weight.ExtraBold))
        self.about_button.setFlat(True)
        self.about_button.setStyleSheet('QPushButton {color: blue; border: none}')
        self.about_button.clicked.connect(self.about)
        self.about_button.setAutoDefault(False)
        self.about_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.about_button.setToolTip("Show the About box, which includes a link to the LTMatch page on the MobileRead forums.")
        button_layout.addWidget(self.about_button)

        # configure
        self.conf_button = QPushButton(_('&Configure...'), self)
        self.conf_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.conf_button.clicked.connect(self.config)
        button_layout.addWidget(self.conf_button)

        # load lt books
        self.load_lt_button = QPushButton(_('&Load LT Books'), self)
        self.load_lt_button.clicked.connect(self.load_lt_books)
        self.load_lt_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        button_layout.addWidget(self.load_lt_button)

        # auto match
        self.auto_match_button = QPushButton(_('&Identify matches...'), self)
        self.auto_match_button.clicked.connect(self.auto_match)
        self.auto_match_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.auto_match_button.setEnabled(False)       # disabled until they "load lt books"
        button_layout.addWidget(self.auto_match_button)

        # toggle show matched
        self.show_matched_check = QCheckBox(_("Show matched"), self)
        self.show_matched_check.setChecked(True)
        self.show_matched_check.stateChanged.connect(self.toggle_show_matched)
        self.show_matched_check.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.show_matched_check.setEnabled(False)       # disabled until they "load lt books"
        button_layout.addWidget(self.show_matched_check)

        # stretch
        button_layout.addStretch(6)

        # match selected
        self.match_selected_button = QPushButton(_('&Match selected'), self)
        self.match_selected_button.clicked.connect(self.match_selected)
        self.match_selected_button.setEnabled(False)       # disabled until they "load lt books"
        button_layout.addWidget(self.match_selected_button, alignment=Qt.AlignHCenter)

        # stretch
        button_layout.addStretch(7)

        # ignore
        self.ignore_button = QPushButton(_('I&gnore'), self)
        self.ignore_button.clicked.connect(self.ignore_selected)
        self.ignore_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.ignore_button.setEnabled(False)      # until user does "load lt books"
        button_layout.addWidget(self.ignore_button)

        # save
        self.save_button = QPushButton(_('&Save all matches...'), self)
        self.save_button.clicked.connect(self.save_matches)
        self.save_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.save_button.setEnabled(False)      # until user does "Identify matches..."
        button_layout.addWidget(self.save_button)

        # update calibre ratings
        self.update_ratings_button = QPushButton(_('&Update ratings...'), self)
        self.update_ratings_button.clicked.connect(self.update_ratings)
        self.update_ratings_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.update_ratings_button.setEnabled(False)      # enabled after "load lt books" IFF custom rating column exists
        button_layout.addWidget(self.update_ratings_button)

        # close
        self.close_button = QPushButton(_('Close'), self)
        self.close_button.clicked.connect(self.close)
        self.close_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        button_layout.addWidget(self.close_button)

        # load the calibre books.
        self.load_calibre_books()

        # Cause our dialog size to be restored from prefs or created on first usage
        self.resize_dialog()


    # these next two methods are purely to display the warning about the missing lt_id custom column
    # AFTER the MatchDialog is shown. Madness!
    def showEvent(self, event):
        super(MatchDialog, self).showEvent(event)
        QTimer.singleShot(1000, self.checkColumns)


    def checkColumns(self):

        if self.lt_id_col not in self.db.custom_field_keys(include_composites=False):
            msgBox = QMessageBox(self)
            msgBox .setWindowTitle("LibraryThing Match")
            msgBox.setText(_("Cannot find custom column ") + self.lt_id_col)
            msgBox.setInformativeText(_("For full functionality, you should create or specify a custom column to store LibraryThing IDs. "
                                      "See the LibraryThing Match documentation for further details."))
            msgBox.setIcon(QMessageBox.Information)
            msgBox.exec_()
            self.lt_id_col = None

        # don't need a message for this - disabled button will be enough
        if self.lt_rating_col not in self.db.custom_field_keys(include_composites=False):
            self.lt_rating_col = None


    def about(self):
        # Get the about text from a language-specific HTML-formatted file inside the plugin zip file
        lang = get_lang()
        if lang:
            lang = lang[:2]
        else:
            lang = 'en'
        logger.info("Language: %s", lang)
        
        # get_resources doesn't raise a catchable exception if the file doesn't exist. So have to do this.
        text = get_resources('about/about_%s.txt'%lang)
        if not text:
            text = get_resources('about/about_en.txt')

        if text:
            QMessageBox.about(self, _("About the LibraryThing Match plugin"), text.decode('utf-8'))


    def ignore_selected(self):

        # get_selected_book() returns a LTMatchBookWidget
        lt_sel = self.lt_table.get_selected_book()

        if lt_sel is None:
            msgBox = QMessageBox(self)
            msgBox.setText("You need to select a LibraryThing book to ignore.")
            msgBox.setIcon(QMessageBox.Information)
            msgBox.exec_()
        else:
            lt_id = lt_sel.getID()
            self.ignore_list.append(lt_id)
            prefs['ignore_list'] = self.ignore_list
            lt_sel.setDisabled()
            row = lt_sel.getListIndex()
            if row < self.lt_table.rowCount()-1:
                self.lt_table.selectRow(row+1)


    def match_selected(self):

        # get_selected_book() returns a LTMatchBookWidget
        cal_itm = self.calibre_table.get_selected_book()
        lt_itm = self.lt_table.get_selected_book()

        if cal_itm is None or lt_itm is None:
            msgBox = QMessageBox(self)
            msgBox.setText(_("You must select exactly one row on each side."))
            msgBox.setIcon(QMessageBox.Information)
            msgBox.exec_()
        else:
            copytoltid = prefs['use_calibre_identifier']
            book_id = cal_itm.getID()
            lt_id = lt_itm.getID()
            mi = self.db.new_api.get_metadata(book_id)
            mi.set(self.lt_id_col, lt_id)
            if copytoltid:
                mi.set_identifier(LTID_NAME, lt_id)
            self.db.new_api.set_metadata(book_id, mi, set_title=False, set_authors=False)

            # matched is a dictionary of key-tuple pairs, where key is lt_id, and tuple is calibre_id, row
            self.matched[lt_id] = (book_id, cal_itm.getListIndex())

            # move the lt_itm into the same row as the cal_itm if they're different.
            # swapsies returns the lt_id of the item we're replacing.
            dest_id = self.swapsies(lt_itm, cal_itm.getListIndex(), _("Matched"))

            # remove BOTH items from possible_match, so we don't overwrite any manual matches
            try:
                self.possible_match.pop(lt_id)
            except KeyError:
                logger.warning(_("match_selected: lt_id key %s not in possible_match dictionary"), lt_id)

            try:
                self.possible_match.pop(dest_id)
            except KeyError:
                logger.warning(_("match_selected: dest_id key %s not in possible_match dictionary"), dest_id)


    def swapsies(self, lt_itm, destination_row, new_status, new_status_percent=0, highlight_selected_row = True):

        # move the lt_itm into the same row as the matching calibre item (if they're different).
        # (the destination row may be blank, or be indicating a match with a missing item.)
        # because it's confusing, it's worth reiterating that this is operating on the same table, i.e. the LT table.
        source_index = lt_itm.getListIndex()

        dest_id = ""

        if source_index != destination_row:

            source_book = self.lt_table.remove_item(source_index)
            source_book.setListIndex(destination_row)

            # destination row may be blank or "Missing"
            dest_book = self.lt_table.remove_item(destination_row)
            if dest_book is not None:
                dest_id = dest_book.getID()
                # dest_id will be None for "Missing" items.
                if dest_id is not None:
                    dest_book.setListIndex(source_index)
                    dest_book.update_item(self.lt_table)
                    self.lt_collection[dest_id]["row_num"] = source_index

            source_book.update_item(self.lt_table)

            source_id = lt_itm.getID()
            self.lt_collection[source_id]["row_num"] = destination_row

            # clear the status from the previous row
            # reminder: status_table is a qTW, so use takeItem()
            self.status_table.takeItem(source_index, 0)

            if highlight_selected_row:
                self.lt_table.selectRow(destination_row)

        # regardless of whether they were in the same row or not, we still need to update the status table
        itm = LTMatchStatusWidget()
        itm.setMatchStatus(new_status, new_status_percent)
        self.status_table.setItem(destination_row, 0, itm)

        return dest_id


    def auto_match(self):
    
        displayed_count = self.m.count()

        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("LibraryThing Match")

        if displayed_count > 0:
            msgBox.setText(_("Identify possible matches?"))
            msgBox.setInformativeText(_("This will use fuzzy matching to identify possible/probable matches between "
                                      "your calibre e-books and LibraryThing collection. It may take some time.\n\n"
                                      "Do you wish to continue?"))
            msgBox.setIcon(QMessageBox.Question)
            msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
            msgBox.setDefaultButton(QMessageBox.Ok)
            ret = msgBox.exec_()
            if ret != QMessageBox.Ok:
                logger.info(_("Clicked Cancel"))
                return 
        else:
            msgBox.setText(_("Identify possible matches requires books to match against."))
            msgBox.setInformativeText(_("You need some calibre e-books AND LibraryThing books to match against."))
            msgBox.setIcon(QMessageBox.Information)
            msgBox.setStandardButtons(QMessageBox.Ok)
            ret = msgBox.exec_()
            return 


        # modal progress dialog
        pd = QProgressDialog(_("Matching operation in progress..."), _("Cancel"), 0, displayed_count, self)
        pd.setWindowModality(Qt.WindowModal)

        # iterate every calibre book in the list
        # using self.book_ids, so that x matches the row number        
        for x in range(displayed_count):

            pd.setValue(x)
            if pd.wasCanceled():
                # considered (and initially used) a break here, but we could have a false half-match.
                return
            book_id = self.book_ids[x]
            mi = self.db.new_api.get_metadata(book_id)
            cal_author = mi.author_sort
            cal_title = mi.title
            if self.lt_id_col is not None and mi.get(self.lt_id_col) is not None:
                continue

            # iterate every LT book in the LT list, looking for matches
            # yes, a nested loop...
            #   There is a "process()" function in fuzzywuzzy, but it's output is not suitable.
            highest_match_prob = 0
            highest_matched_id = None
            for k, v in self.lt_collection.items():
                if k in self.matched or k in self.ignore_list:
                    continue
                lt_author = v["author_lf"]
                lt_title = v["title"]
                # tested the other four algorithms. token_set_ratio() is the best by a fair margin.
                match_prob = fuzz.token_set_ratio(lt_author + " " + lt_title, cal_author + " " + cal_title)
                if match_prob > highest_match_prob:
                    highest_match_prob  = match_prob
                    highest_matched_id = k

            # we could get to here and have no probable matches due to everything already being matched.
            if highest_matched_id is None:
                continue

            # we keep track of the best match in a dict. Sometimes, especially with low-quality matches, but even with 100%s,
            # the same LT book is the best match for several calibre books. in that case, we only replace it here
            # if it's better that the previous match. Example, from top to bottom:
            #   Wilson, Robert Charles Bios is  55% match f/ Brontë, Charlotte Jane Eyre
            #                                   47%       f/ Dickens, Charles David Copperfield
            #                                   52%       f/ Dickens, Charles A Tale of Two Cities 
            #                                  100%       f/ Wilson, Robert Charles Bios
            # so we'd save the first one, ignore the second and third, and overwrite with the fourth.
            
            # self.possible_match is a dict - key:lt_id, value: tuple(calibre_id, row, percent).
            if highest_matched_id in self.possible_match:
                if highest_match_prob > self.possible_match[highest_matched_id][2]:
                    self.possible_match[highest_matched_id] = (book_id, x, highest_match_prob)
            else:
                self.possible_match[highest_matched_id] = (book_id, x, highest_match_prob)


        logger.info (_("Finished fuzzy match analysis." ))

        # need to move the matching LT book into the same row as the corresponding calibre book,
        # luckily, we stored the "row_num" in the lt_collection.
        for k, v in self.possible_match.items():
            lt_row_num = self.lt_collection[k]["row_num"]
            cal_row_num = v[1]
            percent = v[2]
            lt_itm = self.lt_table.item(lt_row_num, 0)
            self.swapsies(lt_itm, cal_row_num, _("Partial"), percent, highlight_selected_row = False)

        logger.info (_("Finished moving books." ))

        # now they're allowed to auto-save (if there's a column for doing it).
        self.save_button.setEnabled(self.lt_id_col is not None)

        pd.setValue(displayed_count)


    def save_matches(self):

        threshold = int(prefs['probable_match_threshold'])
        copytoltid = prefs['use_calibre_identifier']

        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("LibraryThing Match")
        msgBox.setText(_("Save all probable matches?"))
        msgBox.setInformativeText(_("This will save matches for LibraryThing books where the match "
                                  "probability is greater than or equal to your defined threshold value, i.e. {threshold}%.\n\n"
                                  "Do you wish to continue?").format(threshold = threshold))
        msgBox.setIcon(QMessageBox.Question)
        msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        msgBox.setDefaultButton(QMessageBox.Ok)
        ret = msgBox.exec_()
        if ret != QMessageBox.Ok:
            return

        # self.possible_match is a dict - key:lt_id, value: tuple(calibre_id, row, percent).
        for k, v in self.possible_match.items():
            # ignore any we've manually matched since the "Identify" was run.
            if k in self.matched or k in self.ignore_list:
                continue
            lt_id = k
            book_id = v[0]
            row_num = v[1]
            percent = v[2]
            if percent >= threshold:
                mi = self.db.new_api.get_metadata(book_id)
                logger.debug (_("About to save %s %s %s %s"), mi.author_sort, mi.title, lt_id, percent)

                mi.set(self.lt_id_col, lt_id)
                if copytoltid:
                    mi.set_identifier(LTID_NAME, lt_id)
                self.db.new_api.set_metadata(book_id, mi, set_title=False, set_authors=False)

                # update matched list (a dictionary of key-tuple pairs, where key is lt_id, and tuple is (calibre_id, row))
                self.matched[lt_id] = (book_id, row_num)

                # don't need to swap rows coz they're already aligned.

                # status table.
                itm = LTMatchStatusWidget()
                itm.setMatchStatus(_("Matched"))
                self.status_table.setItem(row_num, 0, itm)


    def toggle_show_matched(self):

        # "tableWidget.setRowHidden(row, True) works very well for filtering" said somebody somewhere
        state = self.show_matched_check.checkState() != Qt.Checked

        for k, v in self.matched.items():
            row = v[1]    # row
            self.lt_table.setRowHidden(row, state)
            self.status_table.setRowHidden(row, state)
            self.calibre_table.setRowHidden(row, state)


    def update_ratings(self):

        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("LibraryThing Match")
        msgBox.setText(_("Update ratings for matched books?"))
        msgBox.setInformativeText(_("This will update the specified ratings custom column with your LibraryThing rating for matched books.\n\n"
                                  "Do you wish to continue?"))
        msgBox.setIcon(QMessageBox.Question)
        msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        msgBox.setDefaultButton(QMessageBox.Ok)
        ret = msgBox.exec_()
        if ret != QMessageBox.Ok:
            logger.info (_("Clicked Cancel"))
            return 

        # self.matched is a dict - key:lt_id, value: tuple(calibre_id, row).
        # LT uses half-star ratings, so we need to multiply by two before saving in our custom column,
        # which uses half-star ratings.
        for k, v in self.matched.items():
            # books from the "matched" list might not be in the current lt_collection
            # (if they were previously matched and now we've changed the collection)
            if k in self.lt_collection:
                lt_id = k
                book_id = v[0]
                rating = self.lt_collection[k]["rating"]
                if rating != 0:
                    mi = self.db.new_api.get_metadata(book_id)
                    logger.debug (_("About to rate %s %s %s %s"), mi.author_sort, mi.title, lt_id, rating)
                    mi.set(self.lt_rating_col, 2 * rating)
                    self.db.new_api.set_metadata(book_id, mi, set_title=False, set_authors=False)


    def close(self):

        self.accept()


    def config(self):

        if self.do_user_config(parent=self):
            # maybe a lot of things we could do after prefs set, but updating the caption is definitely one.
            self.update_lt_caption()


    def update_lt_caption(self):

        lt_label = (
                     '<b>LibraryThing</b>' 
                     + ' // ' + prefs['librarything_user'] 
                     + ' // ' + prefs['librarything_coll']
                     +(' // ' + prefs['librarything_taglist'] if prefs['librarything_taglist'] != '' else '')
                   )
                   
        self.lt_label.setText(lt_label)


    def scrolled_lt_list(self):

        # Get the scroll bar from the QTableWidget
        scroll_to = self.lt_table.verticalScrollBar().value()
        self.calibre_table.verticalScrollBar().setValue(scroll_to)
        self.status_table.verticalScrollBar().setValue(scroll_to)


    def scrolled_cal_list(self):

        # Get the scroll bar from the QTableWidget
        scroll_to = self.calibre_table.verticalScrollBar().value()
        self.lt_table.verticalScrollBar().setValue(scroll_to)
        self.status_table.verticalScrollBar().setValue(scroll_to)


    def scrolled_status_list(self):

        # Get the scroll bar from the QTableWidget
        scroll_to = self.status_table.verticalScrollBar().value()
        self.lt_table.verticalScrollBar().setValue(scroll_to)
        self.calibre_table.verticalScrollBar().setValue(scroll_to)


    def load_calibre_books(self):

        try:

            logger.info(_("populate calibre books"))
            tbl = self.calibre_table

            displayed_count = self.m.count()
            rows = list(range(displayed_count))
            self.book_ids = list(map(self.m.id, rows))
            tbl.setRowCount(displayed_count)
            for x in range(displayed_count):
                book_id = self.book_ids[x]
                mi = self.db.new_api.get_metadata(book_id)
                author = mi.author_sort
                title = mi.title
                lt_id = mi.get(self.lt_id_col)  # if the column is missing, self.lt_id_col hasn't been set to None yet, so this is "sort of" okay.
                if lt_id is not None:
                    self.matched[lt_id] = (book_id, x)
                book = LTMatchBookWidget(author, title, index = x, id = book_id)
                book.update_item(tbl)

            # we don't know how many rows LT table (and status table) need until we've loaded LT books.

        except Exception as e:
            logger.error(_("Calibre list population failed: %s"), repr(e))
            msgBox = QMessageBox(self)
            msgBox .setWindowTitle("LibraryThing Match")
            msgBox.setText(_("Can't load calbre e-books to an error."))
            msgBox.setInformativeText(_("See error details below."))
            msgBox.setDetailedText(repr(e))
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.exec_()
            return


    def load_lt_books(self):

        url = "https://www.librarything.com/api_getdata.php?userid={user}&max={max}&tagList={taglist}&showCollections=1&showRatings=1&showTags=1&booksort={order}&responseType=json"
        url = url.format(
                            user = prefs['librarything_user'],
                            max = prefs['max_results'],
                            taglist = urllib.parse.quote_plus(prefs['librarything_taglist']),
                            order = prefs['librarything_sort']
                        )

        try:
            logger.info("load_lt_books: Url = %s", url)
            QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))

            response = urllib.request.urlopen(url)
            logger.info ("msg %s", response.msg)
            logger.info ("status %s", response.status)

            data = json.loads(response.read())
            if data["books"]:
                logger.info(_("Connection successful. Retrieved %s books."), len(data["books"]))

        except Exception as e:
            QApplication.restoreOverrideCursor()
            logger.error(_("Connection failed."))
            msgBox = QMessageBox(self)
            msgBox .setWindowTitle("LibraryThing Match")
            msgBox.setText(_("Can't connect to LibraryThing due to an error."))
            msgBox.setInformativeText(_("This could be due to an incorrect username. See error details below."))
            msgBox.setDetailedText(repr(e))
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.exec_()
            return

        
        try:

            all_lt_books = data["books"]

            # a list of LT IDs to ignore.
            if prefs['ignore_list'] != '':
                self.ignore_list = prefs['ignore_list']
                for k in self.ignore_list:
                    if k in all_lt_books:
                        all_lt_books.pop(k)

            # if the user has specified a collection other than "All collections", we filter results
            lt_coll = prefs['librarything_coll']
            if lt_coll == _("All collections"):
                self.lt_collection = all_lt_books
            else:
                # here, we search for books belonging to the user's preferred collection.
                # (since the LT JSON Books API doesn't allow searching by collection.)
                # copy valid books into local storage.
                self.lt_collection = {}
                for k, v in all_lt_books.items():
                    logger.debug (_("load_lt_books: book %s %s"), k, v["title"])
                    # get another dict containing the collections.
                    coll = v["collections"]
                    # iterate the collection items, checking if book belongs to collection selected by user.
                    # (books can be in no collections. if so, we move on.)
                    if not coll:
                        continue
                    logger.debug (_("    Collections: %s"), list(coll.values()))
                    for k2, v2 in coll.items():
                        if v2 == lt_coll:
                            self.lt_collection[k] = v
                            break

            logger.info (_("LT Collection: %s, %s"), lt_coll, len(self.lt_collection))

            if len(self.lt_collection) == 0:
                QApplication.restoreOverrideCursor()
                msgBox = QMessageBox(self)
                msgBox .setWindowTitle("LibraryThing Match")
                msgBox.setText(_("Connection successful but no books found"))
                msgBox.setInformativeText(_("This can happen if the collection is not exactly the same as it appears in LibraryThing (case sensitive), or "
                                          "Max Results is not high enough to encompass All collections. (The LibraryThing API "
                                          "does not allow searching by collection, so we retrieve all books then filter.)\n\n"
                                          "Try leaving the collection blank and sorting by Entry Date reversed. "
                                          "See the LibraryThing Match documentation for further details."))
                msgBox.setIcon(QMessageBox.Information)
                msgBox.exec_()
                return

            # we need to make sure all three tables match the largest table size.
            # logic did my head in, but it turns out quite simple:
            #   calibre table = number of books = easy peasy
            #   LT table = number of books + number of "missing" matches.
            #   final row count is greater of the two.
            # since we have a list of LT books and a list of matched LT IDs, can use a simple set operation
            # to work out the "missing" matches.

            matched_set = set(self.matched.keys())
            lt_set = set(self.lt_collection.keys())
            diff = matched_set.difference(lt_set)

            max_rows = max(self.m.count(), len(self.lt_collection) + len(diff))

            self.lt_table.setRowCount(max_rows)
            self.status_table.setRowCount(max_rows)
            self.calibre_table.setRowCount(max_rows)

            self.lt_table.clearContents()
            self.status_table.clearContents()

            # use a copy of the books, so we can delete (pop) the matched ones as we add them
            books_to_add = self.lt_collection.copy()

            # first, load the already matched books in the correct row, i.e. same row as their calibre counterpart.
            # "matched" is a dictionary of key-tuple pairs, where key is lt_id, and tuple is (calibre_id, row).
            # depending on how the user has configured the LT connection, matched books may not exist in the LT dictionary,
            # but we hold them in the matched dictionary because they'll be useful when iterating calibre books for auto-match.
            for k, v in self.matched.items():
                ltid = k
                i = v[1]    # row
                # if the LT book is in the current collection, we add it, otherwise add a place-holder.
                # bit of duplication here, but unavoidable if we want to italicize and disable "missing" items.
                if k in books_to_add:
                    author = html.unescape(books_to_add[k]["author_lf"])
                    title = html.unescape(books_to_add[k]["title"])
                    book = LTMatchBookWidget(author, title, index = i, id = ltid)
                    book.update_item(self.lt_table)
                    self.lt_collection[k]["row_num"] = i
                    books_to_add.pop(k)
                else:
                    author = _("(Missing)")
                    title = _("(Book not in currently configured LT collection)")
                    book = LTMatchBookWidget(author, title, index = i)
                    book.setDisabled()
                    book.update_item(self.lt_table)

                # still highlight the matched calibre e-books, so the user doesn't get confused.
                itm = LTMatchStatusWidget()
                itm.setMatchStatus(_("Matched"))
                self.status_table.setItem(i, 0, itm)

            logger.info (_("load_lt_books: finished adding matched books"))

            # add the rest of the books in the first available (empty) row.
            # getFirstEmptyRow starts searching from the next row AFTER the passed index.
            i = -1
            for k, v in books_to_add.items():
                ltid = k
                author = html.unescape(v["author_lf"])
                title = html.unescape(v["title"])
                i = self.lt_table.getFirstEmptyRow(i)
                logger.debug (_("load_lt_books: row %s - %s %s %s"), i, ltid, author, title)

                book = LTMatchBookWidget(author, title, index = i, id = ltid)
                book.update_item(self.lt_table)
                self.lt_collection[k]["row_num"] = i

            logger.info (_("load_lt_books: finished adding other books"))

            # now they're allowed to match, auto-match, ignore, show/hide matched, and update ratings
            # IFF custom columns are available.
            self.auto_match_button.setEnabled(True)
            self.show_matched_check.setEnabled(self.lt_id_col is not None)
            self.match_selected_button.setEnabled(self.lt_id_col is not None)
            self.ignore_button.setEnabled(True)
            self.update_ratings_button.setEnabled(self.lt_id_col is not None and self.lt_rating_col is not None)

            # clear the possible matches dictionary in case we're running this again after already doing "Identify matches"
            # (we don't clear the actual matched items because we want to use them when loading the LT list.)
            self.possible_match = {}


        except Exception as e:
            QApplication.restoreOverrideCursor()
            logger.error(_("LT list loading failed."))
            # "Just printing (e) is cleaner and more likely to work." said somebody somewhere
            msgBox = QMessageBox(self)
            msgBox .setWindowTitle("LibraryThing Match")
            msgBox.setText(_("Can't load LibraryThing books to an error."))
            msgBox.setInformativeText(_("See error details below."))
            msgBox.setDetailedText(repr(e))
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.exec_()
            return

        finally:
            QApplication.restoreOverrideCursor()


    def show_lt_details_popup(self):

        # show the JSON record for this book.
        itm = self.lt_table.get_selected_book()
        if itm is not None and itm.getID() is not None:
            id = itm.getID()
            v = self.lt_collection[id]
            jsontext = json.dumps(v, indent=4)
            jsontext = html.unescape(jsontext)
            ViewDetails(_("LibraryThing Book"), jsontext, parent=self).exec_()


    def show_calibre_details_popup(self):

        # TODO: Use Book Info screen, or convert metadata to html.
        #       might be possible to derive code from class ShowBookDetailsAction(InterfaceAction)
        #       in src\calibre\gui2\actions\show_book_details.py
        itm = self.calibre_table.get_selected_book()
        if itm is not None and itm.getID() is not None:
            id = itm.getID()
            mi = self.gui.current_db.new_api.get_metadata(id)
            ViewDetails(_("calibre e-book"), mi.__str__(), parent=self).exec_()
            #html = render_html(mi, True, None)    #DOESN'T WORK
            #ViewDetails("calibre Book", html, parent=self).exec_()
